{
    "name": "{{ package_name }}",
    "version": "1.0.0",
    "description": "Generated TypeScript classes from Avro schema",
    "main": "dist/index.js",
    "type": "module",
    "scripts": {
        "build": "tsc",
        "start": "node dist/index.js"
    },
    "dependencies": {
        "avro-js": "^1.12.0",
        "typedjson": "^1.8.0",
        "pako": "^2.1.0",
        "reflect-metadata": "^0.2.2",
        "@types/pako": "^2.0.3"
    },
    "devDependencies": {
        "typescript": "^5.6.2"
    },
    "author": "",
    "license": "ISC"
}
