<Project Sdk="Microsoft.NET.Sdk">
    <PropertyGroup>
        <TargetFramework>net9.0</TargetFramework>
        <Nullable>enable</Nullable>
        <GenerateDocumentationFile>true</GenerateDocumentationFile>
    </PropertyGroup>
    <ItemGroup>
        {%- if avro_annotation %}
        <PackageReference Include="Apache.Avro" Version="1.12.0" />
        {%- endif %}
        {%- if newtonsoft_json_annotation %}
        <PackageReference Include="Newtonsoft.Json" Version="13.0.3" />
        {%- endif %}
        {%- if system_text_json_annotation %}
        <PackageReference Include="System.Text.Json" Version="9.0.3" />
        {%- endif %}
        <PackageReference Include="System.Memory.Data" Version="9.0.3" />
    </ItemGroup>
    <ItemGroup>
        <PackageReference Include="NUnit" Version="4.3.2" />
        <PackageReference Include="NUnit3TestAdapter" Version="5.0.0" />
        <PackageReference Include="Microsoft.NET.Test.Sdk" Version="17.13.0" />
    </ItemGroup>
</Project>  