<Project Sdk="Microsoft.NET.Sdk">
    <PropertyGroup>
        <TargetFramework>net9.0</TargetFramework>
        <Nullable>enable</Nullable>
        <GenerateDocumentationFile>true</GenerateDocumentationFile>
    </PropertyGroup>
    <ItemGroup>
        <ProjectReference Include="../src/{{project_name | pascal}}.csproj"/>
    </ItemGroup>
    <ItemGroup>
        <PackageReference Include="NUnit" Version="4.3.2" />
        <PackageReference Include="NUnit3TestAdapter" Version="5.0.0" />
        <PackageReference Include="Microsoft.NET.Test.Sdk" Version="17.13.0" />
    </ItemGroup>
</Project>  